# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#

import itertools
from . import DataSource


class MultiSource(DataSource):
    def __init__(self, *sources):
        if len(sources) == 1 and isinstance(sources[0], list):
            sources = sources[0]

        self.sources = sources
        self._lengths = [None] * len(sources)

    def __iter__(self):
        return itertools.chain(*self.sources)

    def __getitem__(self, n):

        if n < 0:
            n = len(self) + n

        i = 0
        while n >= self._length(i):
            n -= self._length(i)
            i += 1
        return self.sources[i][n]

    def __len__(self):
        return sum(self._length(i) for i, _ in enumerate(self.sources))

    def _length(self, i):
        if self._lengths[i] is None:
            self._lengths[i] = len(self.sources[i])
        return self._lengths[i]

    def to_xarray(self):
        import xarray as xr

        # return xr.concat(s.to_xarray() for s in self.sources)
        return xr.merge(s.to_xarray() for s in self.sources)


source = MultiSource



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#


from docutils import nodes
from docutils.parsers.rst import Directive


class SourcesDirective(Directive):
    def run(self):

        from climetlab.sources import list_entries

        return [nodes.paragraph(text=x) for x in list_entries()]


def setup(app):

    app.add_directive("sources", SourcesDirective)

    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#


import os
import yaml
import climetlab


def _find(collection, yaml_file, fail):
    top = os.path.dirname(climetlab.__file__)
    for path, dirs, files in os.walk(top):
        if os.path.basename(path) == collection:
            if yaml_file in files:
                return os.path.join(path, yaml_file)

    if fail:
        raise Exception("Cannot find '%s' in '%s'" % (yaml_file, collection))


def load(collection, name, fail=True):
    path = _find(collection, name + ".yaml", fail)
    if path is not None:
        with open(path) as f:
            return yaml.load(f.read(), Loader=yaml.SafeLoader)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#

import os
import tempfile
import sys

from climetlab.data import load as load_data

# Examples
# https://github.com/ecmwf/notebook-examples/tree/master/visualisation

from Magics import macro

try:
    from IPython.display import Image, SVG
except Exception:

    def Image(x):
        return x

    def SVG(x):
        return x


class Driver:
    def __init__(self, width=680, grid=False, **kwargs):
        self._projection = None
        self._data = None
        self._format = "png"
        self._width_cm = 10
        self._height_cm = 10
        self._width = width
        self._page_ratio = 1.0
        self._contour = macro.mcont(contour_automatic_setting="ecmwf", legend="off",)

        self._grid = grid
        self._background = macro.mcoast(
            map_grid=self._grid,
            map_grid_colour="tan",
            map_label="off",
            map_boundaries="on",
            map_coastline_land_shade="on",
            map_coastline_land_shade_colour="cream",
            map_coastline_colour="tan",
            map_grid_frame=True,
            map_grid_frame_thickness=5,
        )

        self._foreground = macro.mcoast(
            map_grid=self._grid,
            map_label="off",
            map_grid_frame=True,
            map_grid_frame_thickness=5,
        )

        self._legend = None
        self._title = None

        self.kwargs = dict(**kwargs)

        self.bounding_box(90, -180, -90, 180)

    def bounding_box(self, north, west, south, east):
        self._projection = macro.mmap(
            subpage_upper_right_longitude=float(east),
            subpage_upper_right_latitude=float(north),
            subpage_lower_left_latitude=float(south),
            subpage_lower_left_longitude=float(west),
            subpage_map_projection="cylindrical",
        )
        self._page_ratio = (north - south) / (east - west)

    def plot_grib(self, path, offset):
        self._data = macro.mgrib(
            grib_input_file_name=path,
            grib_file_address_mode="byte_offset",
            grib_field_position=int(offset),
        )

    def plot_netcdf(self, params):
        self._data = macro.mnetcdf(**params)

    def plot_numpy(
        self, data, north, west, south_north_increment, west_east_increment, metadata
    ):
        self._data = macro.minput(
            input_field=data,
            input_field_initial_latitude=float(north),
            input_field_latitude_step=-float(south_north_increment),
            input_field_initial_longitude=float(west),
            input_field_longitude_step=float(west_east_increment),
            input_metadata=metadata,
        )

    def plot_xarray(self, ds, variable, dimension_settings={}):
        self._data = macro.mxarray(
            xarray_dataset=ds,
            xarray_variable_name=variable,
            xarray_dimension_settings=dimension_settings,
        )

    def plot_geopoints(self, path):
        self._data = macro.mgeo(geo_input_file_name=path)
        self.style("red-markers")

    def apply_kwargs(self, kwargs):
        if "style" in kwargs:
            self.style(kwargs.pop("style"))
        if kwargs:
            print("magics.apply_kwargs ignoring", kwargs, file=sys.stderr)

    def _apply(self, collection, value, action, default_attribute=None):

        if value is None:
            return None

        if isinstance(value, dict):
            return action(value)

        if isinstance(value, str):

            data = load_data(collection, value, fail=default_attribute is None)
            if data is None:
                return action({default_attribute: value})

            magics = data["magics"]
            actions = list(magics.keys())
            assert len(actions) == 1, actions

            action = getattr(macro, actions[0])
            return action(magics[actions[0]])

        assert False, (collection, value)

    def projection(self, projection):
        self._projection = self._apply(
            "projections", projection, macro.mmap, "subpage_map_projection"
        )

    def style(self, style):
        self._contour = self._apply("styles", style, macro.mcont)

    def plot_values(self, latitudes, longitudes, values, metadata={}):
        self._data = macro.minput(
            input_type="geographical",
            input_values=list(values),
            input_latitudes_list=list(latitudes),
            input_longitudes_list=list(longitudes),
            input_metadata=metadata,
        )

    def show(
        self, path=None, width=None, title=None, format=None, frame=False, **kwargs
    ):

        if format:
            self._format = format

        if "projection" in self.kwargs:
            self.projection(self.kwargs["projection"])

        if "style" in self.kwargs:
            self.style(self.kwargs["style"])

        tmp = False
        if path is None:
            tmp = True
            fd, path = tempfile.mkstemp("." + self._format)
            os.close(fd)

        _title_height_cm = 0
        if title:
            _title_height_cm = 0.7
            if title is True:
                # Automatic title
                self._title = macro.mtext()
            else:
                self._title = macro.mtext(
                    text_lines=[title],
                    # text_justification='center',
                    # text_font_size=0.6,
                    # text_mode="positional",
                    # text_box_x_position=5.00,
                    # text_box_y_position=18.50,
                    # text_colour='charcoal'
                )

        base, fmt = os.path.splitext(path)
        output = macro.output(
            output_formats=[fmt[1:]],
            output_name_first_page_number="off",
            page_x_length=float(self._width_cm),
            page_y_length=float(self._height_cm) * self._page_ratio,
            super_page_x_length=float(self._width_cm),
            super_page_y_length=float(self._height_cm) * self._page_ratio
            + _title_height_cm,
            subpage_x_length=float(self._width_cm),
            subpage_y_length=float(self._height_cm) * self._page_ratio,
            subpage_x_position=0.0,
            subpage_y_position=0.0,
            output_width=self._width if width is None else width,
            page_frame=frame,
            page_id_line="off",
            output_name=base,
        )

        args = [
            x
            for x in (
                output,
                self._projection,
                self._background,
                self._data,
                self._contour,
                self._foreground,
                self._legend,
                self._title,
            )
            if x is not None
        ]

        macro.plot(*args)

        if self._format == "svg":
            Display = SVG
        else:
            Display = Image

        if tmp:
            img = Display(path, metadata=dict(width=width))
            os.unlink(path)
            return img

        return Display(path)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#

import os
from . import DataSource
from climetlab.core.caching import temp_file
import requests
from tqdm import tqdm
import zipfile
import xarray as xr


class ZipUrl(DataSource):
    def __init__(self, url):
        self.path = temp_file("ZipUrl", url)

        if not os.path.exists(self.path + ".d"):
            if not os.path.exists(self.path + ".zip"):
                print("Downloading", url)
                r = requests.head(url)
                r.raise_for_status()
                size = int(r.headers["content-length"])

                r = requests.get(url, stream=True)
                r.raise_for_status()
                total = 0
                mode = "wb"
                with tqdm(
                    total=size,
                    unit_scale=True,
                    unit_divisor=1024,
                    unit="B",
                    disable=False,
                    leave=False,
                ) as pbar:
                    pbar.update(total)
                    with open(self.path, mode) as f:
                        for chunk in r.iter_content(chunk_size=1024):
                            if chunk:
                                f.write(chunk)
                                total += len(chunk)
                                pbar.update(len(chunk))
                os.rename(self.path, self.path + ".zip")
            print("Unzipping...")
            if not os.path.exists(self.path + ".tmp"):
                os.mkdir(self.path + ".tmp")

            with zipfile.ZipFile(self.path + ".zip", "r") as zip_file:
                for file in tqdm(
                    iterable=zip_file.namelist(),
                    leave=False,
                    total=len(zip_file.namelist()),
                ):
                    zip_file.extract(member=file, path=self.path + ".tmp")

            print("Done...")
            os.rename(self.path + ".tmp", self.path + ".d")
            os.unlink(self.path + ".zip")

        self._xarray = xr.open_mfdataset(
            self.path + ".d/*.nc", combine="by_coords"
        )  # nested

    def to_xarray(self):
        return self._xarray


source = ZipUrl



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#

from climetlab.helpers import helper


# This is needed when running Sphinx on ReadTheDoc
try:
    from .drivers.magics import Driver

except Exception:
    from .drivers.missing import Driver


try:
    from IPython.display import display
except Exception:

    def display(x):
        return x


CURRENT_DRIVER = Driver()


def plot_map(data, *args, **kwargs):
    # This is a standalone plot, so we reset the driver
    CURRENT_DRIVER = Driver(*args, **kwargs)

    if getattr(data, "plot_map", None) is None:
        data = helper(data, *args, **kwargs)

    data.plot_map(CURRENT_DRIVER)

    return display(CURRENT_DRIVER.show(**kwargs))



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#

from .sources import load as load_source
from .datasets import load as load_dataset
from .plotting import plot_map


__version__ = "0.0.37"

import climetlab.source
import climetlab.dataset


from climetlab.sources import DataSource
from climetlab.datasets import Dataset

__all__ = [
    load_source,
    load_dataset,
    plot_map,
    climetlab.dataset,
    climetlab.source,
    Dataset,
    DataSource,
]



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#

# See:
# https://github.com/ecmwf/pdbufr

COLUMNS = ("latitude", "longitude", "data_datetime")


class BUFRReader:
    def __init__(self, path):
        self.path = path

    def to_pandas(self, columns=COLUMNS, filters=None, **kwargs):
        import pdbufr

        return pdbufr.read_bufr(self.path, columns=columns, filters=filters)
